//! Primary interface to working with the Blockfish engine. The [`Bot`] type controls an
//! anytime algorithm that will provide a suggestion for the next move. It may be
//! repeatedly polled by the `think` method in order to attempt to improve the suggestion.

use alloc::boxed::Box;
use alloc::collections::BinaryHeap;
use alloc::vec::Vec;
use core::ops::Range;
use mino::matrix::Mat;
use mino::srs::{Piece, Queue};

mod node;
mod trans;

use crate::bot::node::{Node, RawNodePtr};
use crate::bot::trans::TransTable;
use crate::eval::{features, Features, Rating, Weights};
use crate::Arena;

/// Encompasses an instance of the algorithm.
pub struct Bot {
    evaluator: Evaluator,
    trans: TransTable,
    algorithm: SegmentedAStar,
    // IMPORTANT: `arena` must occur after `algorithm` so that it is dropped last.
    arena: Arena,
    iters: u32,
    metrics: Option<Box<Metrics>>,
}

#[derive(Clone, Eq, PartialEq, Debug)]
pub struct Metrics {
    pub start_features: Features,
    pub start_heuristic: i32,
    pub end_features: Features,
    pub end_heuristic: i32,
    pub end_rating: Rating,
    pub end_iteration: u32,
    // TODO(?) memory usage metrics
    // TODO(?) transposition table metrics
    // TODO(?) A* metrics
}

impl Bot {
    /// Constructs a new bot from the given initial state (matrix and queue).
    pub fn new(weights: &Weights, matrix: &Mat, queue: Queue<'_>) -> Self {
        let arena = bumpalo::Bump::new();
        let root = Node::alloc_root(&arena, matrix, queue);
        let evaluator = Evaluator::new(weights, root);
        let trans = TransTable::new();
        let algorithm = SegmentedAStar::new(root);
        Self {
            evaluator,
            trans,
            algorithm,
            arena,
            iters: 0,
            metrics: None,
        }
    }

    /// Runs the bot, stopping it when it has done `max_iters` iterations since it was
    /// first initialized.  An "iteration" is an abstract unit of work that but should be
    /// roughly proportional to the amount CPU time. Iterations are deterministic, so the
    /// same version of the engine will produce the same suggestions if run the for the
    /// same number of iterations.
    pub fn think(&mut self, max_iters: u32) {
        while self.iters < max_iters {
            let did_update = self.algorithm.step(
                &self.arena,
                &mut self.trans,
                &self.evaluator,
                &mut self.iters,
            );

            if did_update {
                tracing::debug!(
                    "new suggestion @ {}: {:?}",
                    self.iters,
                    self.algorithm.best(),
                );

                if let Some(metrics) = self.metrics.as_deref_mut() {
                    let weights = &self.evaluator.weights;
                    let start = self.algorithm.best();
                    metrics.updated_best(weights, start, self.iters);
                }
            }
        }
    }

    /// Returns the number of iterations done so far.
    pub fn iters(&self) -> u32 {
        self.iters
    }

    /// Return the current best suggested placement. Returns `None` under two possible
    /// conditions:
    /// - `think` has not been called enough times to provide an initial suggestion.
    /// - there are no valid placements for the initial state
    pub fn suggest(&self) -> Option<Piece> {
        self.algorithm.best().root_placement()
    }

    /// Start collecting metrics about the search algorithm. Uses the current suggestion
    /// as the "start" of the metrics data.
    pub fn start_instrumenting(&mut self) {
        let weights = &self.evaluator.weights;
        let start = self.algorithm.best();
        self.metrics = Some(Metrics::new(weights, start, self.iters).into());
    }

    /// Returns the current metrics, if enabled by `start_instrumenting`.
    pub fn metrics(&self) -> Option<Metrics> {
        self.metrics.as_deref().cloned()
    }
}

struct Evaluator {
    weights: Weights,
    root_heuristic: i32,
    root_queue_len: usize,
}

impl Evaluator {
    fn new(weights: &Weights, root: &Node) -> Self {
        Self {
            weights: *weights,
            root_heuristic: features(root.matrix(), 0).evaluate(weights),
            root_queue_len: root.queue().len(),
        }
    }

    fn evaluate(&self, mat: &Mat, queue: Queue<'_>, cleared: &Range<i16>) -> Rating {
        debug_assert!(queue.len() < self.root_queue_len);
        let pcnt = self.root_queue_len - queue.len();

        // check if we cleared the bottom row of the matrix, which is assumed to be the
        // last line of cheese in the race. if so, then consider this node to be a solve
        // and use the piece count as its rating.
        //
        // TODO: make this condition configurable since its a bit of a hack
        if cleared.contains(&0) {
            return Rating::Solve(pcnt as u32);
        }

        let heuristic = features(mat, pcnt).evaluate(&self.weights);
        Rating::Score(self.root_heuristic - heuristic)
    }
}

impl Metrics {
    fn new(weights: &Weights, start: &Node, iters: u32) -> Self {
        let features = features(start.matrix(), 0);
        let heuristic = features.evaluate(weights);
        let rating = start.rating();
        Self {
            start_features: features,
            start_heuristic: heuristic,
            end_features: features,
            end_heuristic: heuristic,
            end_rating: rating,
            end_iteration: iters,
        }
    }

    #[cold]
    fn updated_best(&mut self, weights: &Weights, node: &Node, iters: u32) {
        let features = features(node.matrix(), 0);
        let heuristic = features.evaluate(weights);
        let rating = node.rating();
        self.end_features = features;
        self.end_heuristic = heuristic;
        self.end_rating = rating;
        self.end_iteration = iters;
    }
}

// This implements an algorithm that is very similar to A* but has a slight
// modification. Rather than one big open set, there are separate sets at each depth of
// the search. After picking a node from one open set and expanding its children into the
// successor set, we next pick a node from that successor set. This process continues
// until a terminal node is reached. In order to select which open set to start picking
// from next, we look globally at all the open sets and find the node with the best
// rating; this part works similarly to as if there was only one open set.
//
// Only terminal nodes are compared in order to pick a suggestion. An interesting
// consequence of this design is that on the first run of the algorithm we end up
// performing a best-first-search, and the first terminal node found ends up being our
// initial suggestion. This BFS terminates very quickly so it is nice from the perspective
// of an anytime algorithm.
//
// The problem with directly applying A* for an anytime downstacking algorithm is that
// simply looking for the best heuristic measurement (f) can lead you into a situation
// where a node that only made 2 placements has a better score than all of the nodes with
// 3+ placements, and thus it is considered the best. This is definitely not correct,
// since that 2-placement node only leads to worse board states as you continue to place
// pieces on the board. In downstacking you have to place all of your pieces, you can't
// just stop after placing a few and arriving at a good board state! So before actually
// considering a node to be a suggestion we have to make sure we run out all of the queue
// first (i.e. its a terminal node), and only then should we check its rating.

struct SegmentedAStar {
    open: Vec<BinaryHeap<AStarNode>>,
    depth: usize,
    best: RawNodePtr,
}

impl SegmentedAStar {
    fn new(root: &Node) -> Self {
        let mut open = Vec::with_capacity(root.queue().len());
        open.push(BinaryHeap::new());
        open[0].push(root.into());
        Self {
            open,
            depth: 0,
            best: root.into(),
        }
    }

    fn best(&self) -> &Node {
        unsafe { self.best.as_node() }
    }

    fn step(
        &mut self,
        arena: &Arena,
        trans: &mut TransTable,
        eval: &Evaluator,
        iters: &mut u32,
    ) -> bool {
        *iters += 1;
        match self.expand(arena, trans, eval) {
            Ok(work) => {
                *iters += work;
                false
            }
            Err(maybe_cand) => {
                self.select();
                if let Some(cand) = maybe_cand {
                    self.backup(cand)
                } else {
                    false
                }
            }
        }
    }

    fn expand<'a>(
        &mut self,
        arena: &'a Arena,
        trans: &mut TransTable,
        eval: &Evaluator,
    ) -> Result<u32, Option<&'a Node>> {
        let open_set = self.open.get_mut(self.depth).ok_or(None)?;
        let cand = open_set.pop().ok_or(None)?;
        let cand = unsafe { cand.0.as_node() };

        if cand.is_terminal() {
            return Err(Some(cand));
        }

        self.depth += 1;
        if self.open.len() <= self.depth {
            self.open.resize_with(self.depth + 1, BinaryHeap::new);
        }

        let mut work = 0;
        let evaluate = |mat: &Mat, queue: Queue<'_>, clr: &Range<i16>| {
            // each evaluated board state = +1 unit work
            work += 1;
            eval.evaluate(mat, queue, clr)
        };

        let expanded = cand.expand(arena, trans, evaluate);
        self.open[self.depth].extend(expanded.map(AStarNode::from));

        Ok(work)
    }

    #[cold]
    fn backup(&mut self, cand: &Node) -> bool {
        if cand.rating() > self.best().rating() {
            self.best = cand.into();
            true
        } else {
            false
        }
    }

    fn select(&mut self) {
        let mut best = None::<&Node>;
        self.depth = 0;

        for (depth, set) in self.open.iter().enumerate() {
            let Some(cand) = set.peek() else { continue };
            let cand = unsafe { cand.0.as_node() };
            if best.map_or(true, |best| cand.rating() > best.rating()) {
                best = Some(cand);
                self.depth = depth;
            }
        }
    }
}

// Wraps a `Node` pointer but implements `cmp::Ord` in order to compare by rating.
#[derive(Copy, Clone)]
struct AStarNode(RawNodePtr);

impl From<&Node> for AStarNode {
    fn from(node: &Node) -> Self {
        Self(node.into())
    }
}

impl core::cmp::Ord for AStarNode {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        let lhs = unsafe { self.0.as_node() };
        let rhs = unsafe { other.0.as_node() };
        // TODO: tiebreaker strategy
        lhs.rating().cmp(&rhs.rating())
    }
}

impl core::cmp::PartialOrd for AStarNode {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl core::cmp::Eq for AStarNode {}

impl core::cmp::PartialEq for AStarNode {
    fn eq(&self, other: &Self) -> bool {
        self.cmp(other).is_eq()
    }
}
