//! Transposition table for deduplicating identical states.

// trans rights are human rights :D

use mino::srs::{PieceType, Queue};
use mino::Mat;

use crate::bot::node::{Node, RawNodePtr};
use crate::HashMap;

pub struct TransTable {
    lookup: HashMap<QueueKey, HashMap<RawMatPtr, RawNodePtr>>,
}

impl TransTable {
    /// Constructs a new empty transposition table.
    pub fn new() -> Self {
        Self {
            lookup: HashMap::with_capacity(128),
        }
    }

    /// Looks up a node in the transposition table matching the given state. If one is
    /// found, returns `Err(existing_node_ptr)`. If not found, a node is created and
    /// inserted by calling the given closure, and returns `Ok(newly_inserted_node)`.
    pub fn get_or_insert<'a>(
        &mut self,
        matrix: &Mat,
        queue: Queue<'_>,
        mut alloc: impl FnMut() -> &'a Node,
    ) -> Result<&'a Node, RawNodePtr> {
        // two-phase lookup: first key off queue info, then matrix info

        let map = self
            .lookup
            .entry(QueueKey::new(queue))
            .or_insert_with(HashMap::new);

        // SAFETY: IMPORTANT! this ptr is ONLY used for lookup, NOT insert. it is not
        // necessarily long lived, we are only permitted to insert matrix coming from
        // `alloc()`
        let mat_ptr = unsafe { RawMatPtr::new(matrix) };

        if let Some(&node_ptr) = map.get(&mat_ptr) {
            return Err(node_ptr);
        }

        let node = alloc();
        debug_assert_eq!(*node.matrix(), *matrix);
        debug_assert_eq!(node.queue(), queue);

        // SAFETY: as noted above, this matrix ptr is OK to use for insert, since it comes
        // from the newly allocated node.
        let mat_ptr = unsafe { RawMatPtr::new(node.matrix()) };

        map.insert(mat_ptr, RawNodePtr::from(node));
        Ok(node)
    }
}

// Rather than storing the actual next pieces of each queue, only store the number of
// next pieces. This is because the suffixes will always be the same if they have the
// same number of pieces, so its more efficient to only store & compare the length.
#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
struct QueueKey(u16, Option<PieceType>);

impl QueueKey {
    fn new(queue: Queue<'_>) -> Self {
        Self(queue.next.len() as u16, queue.hold)
    }
}

// share matrix data with the nodes themselves, since the nodes will outlive the
// transposition table.
#[derive(Copy, Clone, Debug)]
#[repr(transparent)]
struct RawMatPtr(*const Mat);

impl RawMatPtr {
    // SAFETY: caller must guaruntee that `mat` lives longer than this value.
    unsafe fn new(mat: &Mat) -> Self {
        Self(mat)
    }

    fn as_mat(&self) -> &Mat {
        unsafe { &*self.0 }
    }
}

impl core::cmp::PartialEq for RawMatPtr {
    fn eq(&self, other: &Self) -> bool {
        *self.as_mat() == *other.as_mat()
    }
}

impl core::cmp::Eq for RawMatPtr {}

impl core::hash::Hash for RawMatPtr {
    fn hash<H: core::hash::Hasher>(&self, h: &mut H) {
        self.as_mat().hash(h)
    }
}
