//! Config file schema (serde) for tidepool. See `example-configs/` for usage samples.

use fish::eval::Weights;
use serde::{Deserialize, Serialize};

#[derive(Deserialize, PartialEq, Debug, Clone)]
pub struct Config {
    pub game: GameConfig,
    #[serde(default)]
    pub bot: BotConfig,
}

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
pub struct GameConfig {
    pub goal: usize,
    #[serde(
        default = "defaults::rules",
        deserialize_with = "de::game_rules_config"
    )]
    pub rules: GameRulesConfig,
}

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
pub struct GameRulesConfig {
    #[serde(default = "defaults::min_garbage")]
    pub min: usize,
    #[serde(default = "defaults::max_garbage")]
    pub max: usize,
    #[serde(default = "defaults::previews")]
    pub previews: usize,
}

impl GameRulesConfig {
    pub const JSTRIS: Self = Self {
        min: 3,
        max: 9,
        previews: 5,
    };
}

#[derive(Deserialize, PartialEq, Debug, Clone)]
pub struct BotConfig {
    // TODO: algorithm
    // TODO: capabililties
    #[serde(default = "defaults::iters")]
    pub iters: u32,
    #[serde(default = "defaults::weights", deserialize_with = "de::weights")]
    pub weights: Weights,
}

impl BotConfig {
    pub const DEFAULT: Self = Self {
        iters: 50_000,
        weights: Weights::DEFAULT,
    };
}

impl Default for BotConfig {
    fn default() -> Self {
        Self::DEFAULT
    }
}

mod defaults {
    use super::*;

    pub const fn min_garbage() -> usize {
        GameRulesConfig::JSTRIS.min
    }

    pub const fn max_garbage() -> usize {
        GameRulesConfig::JSTRIS.max
    }

    pub const fn previews() -> usize {
        GameRulesConfig::JSTRIS.previews
    }

    pub const fn rules() -> GameRulesConfig {
        GameRulesConfig::JSTRIS
    }

    pub const fn iters() -> u32 {
        BotConfig::DEFAULT.iters
    }

    pub const fn weights() -> Weights {
        BotConfig::DEFAULT.weights
    }
}

mod de {
    use super::*;

    pub fn game_rules_config<'de, D>(deserializer: D) -> Result<GameRulesConfig, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        #[derive(Deserialize)]
        #[serde(untagged)]
        enum Variant {
            Preset(Preset),
            Custom(GameRulesConfig),
        }

        #[derive(Deserialize)]
        enum Preset {
            #[serde(rename = "jstris")]
            Jstris,
        }

        Variant::deserialize(deserializer).map(|var| match var {
            Variant::Preset(Preset::Jstris) => GameRulesConfig::JSTRIS,
            Variant::Custom(c) => c,
        })
    }

    pub fn weights<'de, D>(deserializer: D) -> Result<Weights, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        <[i32; 4] as Deserialize>::deserialize(deserializer).map(Weights)
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_de_1() {
        assert_eq!(
            toml::from_str::<Config>(
                "
[game]
goal = 100

[bot]
iters = 5_000
weights = [1, 2, 3, 4]
"
            )
            .unwrap(),
            Config {
                game: GameConfig {
                    goal: 100,
                    rules: GameRulesConfig::JSTRIS,
                },
                bot: BotConfig {
                    iters: 5_000,
                    weights: Weights([1, 2, 3, 4]),
                }
            }
        );
    }

    #[test]
    fn test_de_2() {
        assert_eq!(
            toml::from_str::<Config>(
                "
game.goal = 100
game.rules = { min = 4, previews = 8 }
bot.iters = 7000
"
            )
            .unwrap(),
            Config {
                game: GameConfig {
                    goal: 100,
                    rules: GameRulesConfig {
                        min: 4,
                        max: 9,
                        previews: 8,
                    },
                },
                bot: BotConfig {
                    iters: 7000,
                    weights: Weights::DEFAULT,
                },
            }
        );
    }

    #[test]
    fn test_de_3() {
        assert_eq!(
            toml::from_str::<Config>(
                "
[game]
goal = 18
rules = \"jstris\"
"
            )
            .unwrap(),
            Config {
                game: GameConfig {
                    goal: 18,
                    rules: GameRulesConfig::JSTRIS,
                },
                bot: BotConfig::default()
            }
        );
    }
}
